#!/usr/bin/env python3
import asyncio
import functools
import logging
import os.path
import sys
from typing import (
    List,
    Optional,
)

import fire
import humanfriendly
from stc_geck.advices import BaseDocumentHolder
from termcolor import colored

from .client import StcGeck
from .exceptions import IpfsConnectionError


def exception_handler(func):
    @functools.wraps(func)
    async def wrapper_func(self, *args, **kwargs):
        try:
            return await func(self, *args, **kwargs)
        except IpfsConnectionError as e:
            print(
                f"{colored('ERROR', 'red')}: Cannot connect to IPFS: {e.info}\n"
                f"{colored('HINT', 'yellow')}: Install IPFS to your computer: "
                f"https://docs.ipfs.tech/install/ipfs-desktop/\n"
                f"{colored('HINT', 'yellow')}: Also, ensure IPFS is launched\n"
                f"{colored('HINT', 'yellow')}: Otherwise, you can pass IPFS address of the working instance with "
                f"`--ipfs-http-base-url` parameter: `geck --ipfs-http-base-url http://127.0.0.1:8080 - serve`",
                file=sys.stderr,
            )
            await self.geck.stop()
            raise
    return wrapper_func


class StcGeckCli:
    def __init__(
        self,
        ipfs_http_base_url: str,
        ipfs_api_base_url: str,
        ipfs_data_directory: str,
        grpc_api_endpoint: str,
        index_alias: str,
        timeout: int,
    ):
        self.geck = StcGeck(
            ipfs_http_base_url=ipfs_http_base_url,
            ipfs_api_base_url=ipfs_api_base_url,
            ipfs_data_directory=ipfs_data_directory,
            grpc_api_endpoint=grpc_api_endpoint,
            index_alias=index_alias,
            timeout=timeout,
        )
        self.grpc_api_endpoint = grpc_api_endpoint
        self.index_alias = index_alias

    def prompt(self):
        if self.geck.is_embed:
            print(f"{colored('INFO', 'green')}: Setting up indices...", file=sys.stderr)
        else:
            print(f"{colored('INFO', 'green')}: Using existent instance on {self.geck.grpc_api_endpoint}", file=sys.stderr)

    @exception_handler
    async def documents(self, query_filter: Optional[dict] = None, fields: Optional[List[str]] = None):
        """
        Stream all STC chunks.

        :return: metadata records
        """
        self.prompt()
        async with self.geck as geck:
            async for document in geck.get_summa_client().documents(
                self.index_alias,
                query_filter=query_filter,
                fields=fields,
            ):
                print(document)

    @exception_handler
    async def download(self, query: str, output_path: str):
        """
        Download file from STC using default Summa match queries.
        Examples: `doi:10.1234/abc, isbns:9781234567890`

        :param query: query in Summa match format
        :param output_path: filepath for writing file

        :return: file if record has corresponding CID
        """
        async with self.geck:
            results = await self._search(query)
            output_path, output_path_ext = os.path.splitext(output_path)
            output_path_ext = output_path_ext.lstrip('.')
            if results:
                document_holder = BaseDocumentHolder(results[0])
                print(f"{colored('INFO', 'green')}: Found {query}", file=sys.stderr)
                if link := (
                    document_holder.get_links().get_first_link(output_path_ext)
                    or document_holder.get_links().get_first_link()
                ):
                    print(f"{colored('INFO', 'green')}: Receiving file {query}...", file=sys.stderr)
                    if (real_extension := link.get('extension', 'pdf')) != output_path_ext:
                        print(
                            f"{colored('WARN', 'yellow')}: Receiving file extension `{real_extension}` "
                            f"is not matching with your output path extension `{output_path_ext}`. Changed to correct one.",
                            file=sys.stderr
                        )
                        output_path_ext = real_extension
                    data = await self.geck.download(link['cid'])
                    final_file_name = output_path + '.' + output_path_ext
                    with open(final_file_name, 'wb') as f:
                        f.write(data)
                        f.close()
                        print(f"{colored('INFO', 'green')}: File {final_file_name} is written", file=sys.stderr)
                else:
                    print(f"{colored('ERROR', 'red')}: Not found CID for {query} and extension {output_path_ext}", file=sys.stderr)
            else:
                print(f"{colored('ERROR', 'red')}: Not found {query}", file=sys.stderr)

    @exception_handler
    async def random_cids(self, n: Optional[int] = None, space: Optional[str] = None):
        """
        Return random CIDs for pinning from the STC Hub API.

        :param n: number of items to pin
        :param space: approximate disk space you would like to allocate for pinning
        """
        if not n and not space:
            raise ValueError("`n` or `space_bytes` should be set")
        if space:
            # 3.61MB is the average size of the item
            n = int(humanfriendly.parse_size(space) / (3.61 * 1024 * 1024))
        self.prompt()
        async with self.geck as geck:
            return await geck.random_cids(n=n)

    async def _search(self, query: str, limit: int = 1, offset: int = 0):
        logging.getLogger('statbox').info({'action': 'search', 'query': query})
        summa_client = self.geck.get_summa_client()
        search_request = {
            'index_alias': self.index_alias,
            'query': {'match': {'value': query.lower()}},
            'collectors': [{'top_docs': {'limit': limit, 'offset': offset}}],
            'is_fieldnorms_scoring_enabled': False,
        }
        return await summa_client.search_documents(search_request)

    @exception_handler
    async def search(self, query: str, limit: int = 1, offset: int = 0):
        """
        Searches in STC using default Summa match queries.
        Examples: `doi:10.1234/abc, isbns:9781234567890, "fetal hemoglobin"`

        :param query: query in Summa match format
        :param limit: how many results to return, higher values incurs LARGE performance penalty.
        :param offset:

        :return: metadata records
        """
        self.prompt()
        async with self.geck:
            print(f"{colored('INFO', 'green')}: Searching {query}...", file=sys.stderr)
            return await self._search(query, limit=limit, offset=offset)

    @exception_handler
    async def serve(self):
        """
        Start serving Summa
        """
        self.prompt()
        async with self.geck:
            print(f"{colored('INFO', 'green')}: Serving on {self.grpc_api_endpoint}", file=sys.stderr)
            while True:
                await asyncio.sleep(5)

    @exception_handler
    async def create_ipfs_directory(
        self,
        output_car: str,
        query: str = None,
        limit: int = 100,
        name_template: str = '{id}.{extension}',
    ):
        """
        Stream all STC chunks.

        :return: metadata records
        """
        self.prompt()
        async with self.geck as geck:
            return await geck.create_ipfs_directory(output_car, query, limit, name_template)


async def stc_geck_cli(
    ipfs_http_base_url: str = 'http://127.0.0.1:8080',
    ipfs_api_base_url: str = 'http://127.0.0.1:5001',
    ipfs_data_directory: str = '/ipns/libstc.cc/data',
    grpc_api_endpoint: str = '127.0.0.1:10082',
    index_alias: str = 'nexus_science',
    timeout: int = 120,
    debug: bool = False,
):
    """
    :param ipfs_http_base_url: IPFS HTTP Endpoint
    :param ipfs_api_base_url: IPFS HTTP API Endpoint
    :param ipfs_data_directory: path to the directory with index
    :param grpc_api_endpoint: port used for Summa
    :param index_alias: default index alias
    :param timeout: timeout for requests to IPFS
    :param debug: add debugging output
    :return:
    """
    logging.basicConfig(stream=sys.stdout, level=logging.INFO if debug else logging.ERROR)
    stc_geck_client = StcGeckCli(
        ipfs_http_base_url=ipfs_http_base_url,
        ipfs_api_base_url=ipfs_api_base_url,
        ipfs_data_directory=ipfs_data_directory,
        grpc_api_endpoint=grpc_api_endpoint,
        index_alias=index_alias,
        timeout=timeout,
    )
    return {
        'create-ipfs-directory': stc_geck_client.create_ipfs_directory,
        'documents': stc_geck_client.documents,
        'download': stc_geck_client.download,
        'random-cids': stc_geck_client.random_cids,
        'search': stc_geck_client.search,
        'serve': stc_geck_client.serve,
    }


def main():
    fire.Fire(stc_geck_cli, name='geck')


if __name__ == '__main__':
    main()



from collections import OrderedDict
from typing import (
    List,
    Optional,
)

from aiosumma import SummaClient

TEMPORAL_RANKING_FORMULA = "original_score * custom_score * fastsigm(abs(now - issued_at) / (86400 * 3) + 5, -1)"
PR_TEMPORAL_RANKING_FORMULA = f"{TEMPORAL_RANKING_FORMULA} * fastsigm(iqpr(quantized_page_rank), 0.15)"


default_field_aliases = {
    'ark_id': 'id.ark_ids',
    'ark_ids': 'id.ark_ids',
    'author': 'authors.family',
    'authors': 'authors.family',
    'cid': 'links.cid',
    'doi': 'id.dois',
    'ev': 'metadata.event.name',
    'extension': 'links.extension',
    'format': 'links.extension',
    'isbn': 'metadata.isbns',
    'isbns': 'metadata.isbns',
    'issn': 'metadata.issns',
    'issns': 'metadata.issns',
    'lang': 'languages',
    'manualslib_id': 'id.manualslib_id',
    'nexus_id': 'id.nexus_id',
    'pmid': 'id.pubmed_id',
    'pub': 'metadata.publisher',
    'pubmed_id': 'id.pubmed_id',
    'rd': 'references.doi',
    'ser': 'metadata.series',
    'wiki': 'id.wiki',
}


default_term_field_mapper_configs = {
    'doi': {'fields': ['id.dois']},
    'doi_isbn': {'fields': ['metadata.isbns']},
    'isbn': {'fields': ['metadata.isbns']},
}


default_field_boosts = {
    'authors': 1.7,
    'content': 0.65,
    'extra': 1.7,
    'title': 1.85,
}


async def get_documents_on_topic(
    summa_client: SummaClient,
    topic: str,
    documents: int = 20,
    index_alias: str = 'nexus_science',
    ranking_formula: Optional[str] = None,
    is_fieldnorms_scoring_enabled: bool = False,
    default_fields: List[str] = ("abstract", "title", "content"),
):
    return await summa_client.search_documents({
        "index_alias": index_alias,
        "query": {"boolean": {"subqueries": [{
            "occur": "should",
            "query": {
                "match": {
                    "value": topic,
                    "query_parser_config": {
                        "default_fields": list(default_fields),
                        "field_aliases": default_field_aliases,
                        "field_boosts": default_field_boosts,
                    },
                }
            },
        }]}},
        "collectors": [{"top_docs": {"limit": documents, "scorer": {
            "eval_expr": ranking_formula or PR_TEMPORAL_RANKING_FORMULA,
        }}}],
        "is_fieldnorms_scoring_enabled": is_fieldnorms_scoring_enabled,
    })


def get_full_query_parser_config(query_language: Optional[str] = None, exact_matches_promoter_slop: int = 4, exact_matches_promoter_boost: float = 2.0):
    query_parser_config = {
        'default_fields': ['abstract', 'concepts', 'content', 'extra', 'title'],
        'term_limit': 20,
        'field_aliases': default_field_aliases,
        'field_boosts': default_field_boosts,
        'exact_matches_promoter': {
            'slop': exact_matches_promoter_slop,
            'boost': exact_matches_promoter_boost,
            'fields': ['abstract', 'extra', 'title']
        }
    }
    if query_language:
        query_parser_config['query_language'] = query_language
    return query_parser_config


def get_light_query_parser_config(query_language: Optional[str] = None):
    query_parser_config = {
        'default_fields': ['abstract', 'title'],
        'term_limit': 20,
        'field_aliases': default_field_aliases,
        'field_boosts': default_field_boosts,
    }
    if query_language:
        query_parser_config['query_language'] = query_language
    return query_parser_config


def get_query_parser_config(profile: str, query_language: Optional[str] = None, exact_matches_promoter_boost: float = 2.0):
    if profile == 'full':
        return get_full_query_parser_config(query_language, exact_matches_promoter_boost)
    elif profile == 'light':
        return get_light_query_parser_config(query_language)
    else:
        raise ValueError("Unknown profile")


def get_default_scorer(profile: str):
    if profile == 'full':
        return {'eval_expr': PR_TEMPORAL_RANKING_FORMULA}
    elif profile == 'light':
        return None
    else:
        raise ValueError("Unknown profile")


def format_document(document: dict):
    parts = []
    if title := document.get('title'):
        parts.append(f'Title: {title}')
    if authors := document.get('authors'):
        parts.append(f'Authors: {authors}')
    if id_ := document.get('id'):
        parts.append(f'ID: {id_}')
    if links := document.get('links'):
        parts.append(f'Links: {links}')
    if abstract := document.get('abstract'):
        parts.append(f'Abstract: {abstract[:200]}')
    return '\n'.join(parts)


class BaseDocumentHolder:
    def __init__(self, document):
        self.document = document

    def __getattr__(self, name):
        if name in self.document:
            return self.document[name]
        if name == 'content':
            return
        elif 'metadata' in self.document and name in self.document['metadata']:
            return self.document['metadata'][name]
        elif 'id' in self.document and name in self.document['id']:
            return self.document['id'][name]

    def has_cover(self):
        return bool(self.isbns and len(self.isbns) > 0)

    def get_links(self):
        if self.has_field('links') and self.links:
            return LinksWrapper(self.links)
        return LinksWrapper([])

    @property
    def ordered_links(self):
        links = self.get_links()
        pdf_link = None
        epub_link = None
        other_links = []
        for link in links.links.values():
            if link['extension'] == 'pdf' and not pdf_link:
                pdf_link = link
            elif link['extension'] == 'pdf' and not epub_link:
                epub_link = link
            else:
                other_links.append(link)
        if epub_link:
            other_links = [epub_link] + other_links
        if pdf_link:
            other_links = [pdf_link] + other_links
        return other_links

    @property
    def doi(self):
        if self.has_field('dois') and self.dois:
            return self.dois[0]

    def has_field(self, name):
        return (
            name in self.document
            or name in self.document.get('metadata', {})
            or name in self.document.get('id', {})
        )

    def get_external_id(self):
        if self.doi:
            return f'id.dois:{self.doi}'
        elif self.internal_iso:
            return f'id.internal_iso:{self.internal_iso}'
        elif self.internal_bs:
            return f'id.internal_bs:{self.internal_bs}'
        elif self.pubmed_id:
            return f'id.pubmed_id:{self.pubmed_id}'
        elif self.ark_ids:
            return f'id.ark_ids:{self.ark_ids[-1]}'
        elif self.libgen_ids:
            return f'id.libgen_ids:{self.libgen_ids[-1]}'
        elif self.zlibrary_ids:
            return f'id.zlibrary_ids:{self.zlibrary_ids[-1]}'
        elif self.wiki:
            return f'id.wiki:{self.wiki}'
        elif self.manualslib_id:
            return f'id.manualslib_id:{self.manualslib_id}'
        else:
            return None

    def get_id(self):
        if self.nexus_id:
            return f'id.nexus_id:{self.nexus_id}'
        else:
            return self.get_external_id()


class LinksWrapper:
    def __init__(self, links):
        self.links = OrderedDict()
        for link in links:
            self.add(link)

    def reset(self):
        self.links = OrderedDict()

    def to_list(self):
        return list(self.links.values())

    def add(self, link):
        old_link = self.links.pop(link['cid'], None)
        self.links[link['cid']] = link
        return old_link

    def prepend(self, link):
        old_link = self.add(link)
        self.links.move_to_end(link['cid'], False)
        return old_link

    def delete_links_with_extension(self, extension):
        for cid in list(self.links.keys()):
            if self.links[cid]['extension'] == extension:
                del self.links[cid]

    def override_link_with_extension(self, new_link):
        self.delete_links_with_extension(new_link['extension'])
        return self.prepend(new_link)

    def remove_cid(self, cid):
        return self.links.pop(cid, None)

    def get_link_with_extension(self, extension: Optional[str] = None, from_end: bool = False):
        return self.get_first_link(extension=extension, from_end=from_end)

    def get_first_link(self, extension: Optional[str] = None, from_end: bool = False):
        links = list(self.links.values())
        if from_end:
            links = reversed(links)
        for link in links:
            if extension is None or link['extension'] == extension:
                return link



from izihawa_utils.exceptions import BaseError


class IpfsConnectionError(BaseError):
    pass


class ItemNotFound(BaseError):
    def __init__(self, query):
        self.query = query


class CidNotFound(BaseError):
    def __init__(self, query):
        self.query = query
